/* 10/26/04 - major rehaul 050704

This is an older version - it does the onBoot actions also when the computer wakes up from sleep. A newer
version, ServerWatcher (050903), uses a buffer as test to see if it is still the same server that is waking
up so that it does not load or allocate buffers whenever the computer wakes up.

ServerWatcher registers actions (functions) that an object may want to execute every time that a given server
is booted, or every time that a server quits. If the SC lang client starts up while that server is already on
(was not stopped when quitting the SuperCollider client application), then any onBoot actions will be
executed as soon as the running seårver is recognized by the clients polling mechanism that updates the
default server windows. This means, any classes that want to initialize data on the server as soon as the
server is running will work in both cases:

- When the server is booted from the client
- When the client starts up but the server is already running

Example:

Make object integer 1 respond to Server.local boot and quit:

1.onBoot({ |me| [me, "Ready for action"].postln; });
1.onQuit({ |me| [me, "bye bye"].postln; });

// stop responding to server boots and quits:
1.removeServer;

Same thing using ServerWatcher directly:

ServerWatcher.onBoot(1, { |me, server| [me, "my server booted"].postln; });
ServerWatcher.onQuit(1, { |me, server| [me, "my server quit"].postln; });

*/

SimpleServerWatcher {
	classvar <all;
	var <server;
	var <bootObjects, <quitObjects;

	*initClass {
		all = IdentityDictionary.new;
	}

	*onBoot { | anObject, anAction, aServer |
		this.for(aServer).onBoot(anObject, anAction);
	}

	*onQuit { | anObject, anAction, aServer |
		this.for(aServer).onQuit(anObject, anAction);
	}

	onBoot { | anObject, anAction |
		bootObjects[anObject] = bootObjects[anObject].add(anAction);
		if (server.isRunning) { anAction.(anObject) };
	}

	onQuit { | anObject, anAction |
		quitObjects[anObject] = quitObjects[anObject].add(anAction);
	}

	*for { | aServer |
		var instance;
		aServer = aServer ?? { Server.local };
		instance = all.at(aServer);
		if (instance.isNil) {
			instance = this.new(aServer);
			all.put(aServer, instance);
		};
		^instance;
	}

	*new { | server |
		^super.new.init(server);
	}

	init { |argServer|
		server = argServer ?? { Server.local };
		server.addDependant(this);
		this.clear;
	}

	// remove anObject from both boot and quit list of aServers ServerWatcher
	// if aServer is nil, remove it from all servers. This way
	// if a view does not remember which server it is watching, it can still remove itself
	*remove { | anObject, aServer |
		(all.at(aServer) ?? { all.values }).asArray.do { | sw |
			sw.remove(anObject);
		};
	}

	remove { | anObject |
		bootObjects.removeAt(anObject);
		quitObjects.removeAt(anObject);
	}

	*clear { |server|
		var who;
		if ((who = all.at(server)).notNil) { who.clear };
	}

	clear {
		bootObjects = IdentityDictionary.new;
		quitObjects = IdentityDictionary.new;
	}

	update { | theServer, statusMessage |
		if (statusMessage == \serverRunning) {
			if (theServer.isRunning) {
				bootObjects.keysValuesDo { | object, actions |
					actions.do( _.value(object) );
				}
			}{
				quitObjects.keysValuesDo { | object, actions |
					actions.do( _.value(object) );
				}
			}
		};
	}

}
